#include <windows.h>

#define OHBEXPORT __declspec(dllexport)
#include "ohbbhlp.h"
#include "oheztr.h"

#define OHBINITLOCK() InitializeCriticalSection(&CriticalSection)
#define OHBLOCK() EnterCriticalSection(&CriticalSection)
#define OHBUNLOCK() LeaveCriticalSection(&CriticalSection)
#define OHBDESTROYLOCK() DeleteCriticalSection(&CriticalSection)

#define OHB_HOME_DIR_LEN 2048
#define OHB_EZTR_INIT_STR "CSUSER123455"

typedef struct _OHBEZTR
{
  CHAR szHomeDir[OHB_HOME_DIR_LEN];
  HMODULE hJKMod;
  HMODULE hKJMod;
  int (__stdcall *J2K_InitializeEx)(const char *initStr, const char *homeDir);
  char * (__stdcall *J2K_TranslateMMNT)(int data0, const char *jpStr);
  int (__stdcall *J2K_FreeMem)(char *krStr);
  int (__stdcall *J2K_StopTranslation)(int data0);
  int (__stdcall *J2K_Terminate)(void);
  int (__stdcall *K2J_InitializeEx)(const char *initStr, const char *homeDir);
  char * (__stdcall *K2J_TranslateMMNT)(int data0, const char *krStr);
  int (__stdcall *K2J_FreeMem)(char *jpStr);
  int (__stdcall *K2J_StopTranslation)(int data0);
  int (__stdcall *K2J_Terminate)(void);
} OHBEZTR;

static OHBEZTR eztr;
static CRITICAL_SECTION CriticalSection;
static int bInit;

static int OhbEZTRInit(void);
static int OhbEZTRDestroy(void);
static int OhbInitHomeDir(void);
static int OhbDestroyHomeDir(void);
static int OhbInitEZTR(void);
static int OhbDestroyEZTR(void);

static int OhbEZTRInit(void)
{
  OHBINITLOCK();
  if (OhbInitHomeDir() != 0)
  {
    return -1;
  }
  if (OhbInitEZTR() != 0)
  {
    return -2;
  }

  return 0;
}

static int OhbEZTRDestroy(void)
{
  OhbDestroyEZTR();
  OhbDestroyHomeDir();
  OHBDESTROYLOCK();

  return 0;
}

static int OhbInitEZTR(void)
{
  CHAR szDllBuff[OHB_HOME_DIR_LEN];

  lstrcpyA(szDllBuff, eztr.szHomeDir);
  lstrcatA(szDllBuff, "\\..\\J2KEngine.dll");
  eztr.hJKMod = LoadLibraryA(szDllBuff);
  if (eztr.hJKMod == NULL)
  {
    return -1;
  }

  lstrcpyA(szDllBuff, eztr.szHomeDir);
  lstrcatA(szDllBuff, "\\..\\K2JEngine.dll");
  eztr.hKJMod = LoadLibraryA(szDllBuff);
  if (eztr.hKJMod == NULL)
  {
    return -2;
  }

  eztr.J2K_InitializeEx = (LPVOID)GetProcAddress(eztr.hJKMod,
    "J2K_InitializeEx");
  eztr.J2K_TranslateMMNT = (LPVOID)GetProcAddress(eztr.hJKMod,
    "J2K_TranslateMMNT");
  eztr.J2K_FreeMem = (LPVOID)GetProcAddress(eztr.hJKMod, "J2K_FreeMem");
  eztr.J2K_StopTranslation = (LPVOID)GetProcAddress(eztr.hJKMod,
    "J2K_StopTranslation");
  eztr.J2K_Terminate = (LPVOID)GetProcAddress(eztr.hJKMod, "J2K_Terminate");
  eztr.K2J_InitializeEx = (LPVOID)GetProcAddress(eztr.hKJMod,
    "K2J_InitializeEx");
  eztr.K2J_TranslateMMNT = (LPVOID)GetProcAddress(eztr.hKJMod,
    "K2J_TranslateMMNT");
  eztr.K2J_FreeMem = (LPVOID)GetProcAddress(eztr.hKJMod, "K2J_FreeMem");
  eztr.K2J_StopTranslation = (LPVOID)GetProcAddress(eztr.hKJMod,
    "K2J_StopTranslation");
  eztr.K2J_Terminate = (LPVOID)GetProcAddress(eztr.hKJMod, "K2J_Terminate");

  if (eztr.J2K_InitializeEx == NULL || eztr.J2K_TranslateMMNT == NULL ||
    eztr.J2K_FreeMem == NULL || eztr.J2K_StopTranslation == NULL ||
    eztr.J2K_Terminate == NULL || eztr.K2J_InitializeEx == NULL ||
    eztr.K2J_TranslateMMNT == NULL || eztr.K2J_FreeMem == NULL ||
    eztr.K2J_StopTranslation == NULL || eztr.K2J_Terminate == NULL)
  {
    return -3;
  }

  eztr.J2K_InitializeEx(OHB_EZTR_INIT_STR, eztr.szHomeDir);
  eztr.K2J_InitializeEx(OHB_EZTR_INIT_STR, eztr.szHomeDir);

  return 0;
}

static int OhbDestroyEZTR(void)
{
  if (eztr.J2K_StopTranslation && eztr.K2J_StopTranslation &&
    eztr.J2K_Terminate && eztr.K2J_Terminate)
  {
    eztr.J2K_StopTranslation(0);
/*    eztr.J2K_Terminate(); */
    eztr.K2J_StopTranslation(0);
/*    eztr.K2J_Terminate(); */
  }

  if (eztr.hJKMod)
  {
/*    FreeLibrary(eztr.hJKMod); */
  }

  if (eztr.hKJMod)
  {
/*    FreeLibrary(eztr.hKJMod); */
  }

  ZeroMemory(&eztr, sizeof(eztr));

  return 0;
}

static int OhbInitHomeDir(void)
{
  HKEY hKey;
  DWORD dwLen = OHB_HOME_DIR_LEN;
  int len;

  if (RegOpenKeyExA(HKEY_CURRENT_USER, "Software\\ChangShin\\ezTrans", 0,
    KEY_QUERY_VALUE, &hKey) != ERROR_SUCCESS)
  {
    return -1;
  }

  if (RegQueryValueExA(hKey, "FilePath", NULL, NULL, (LPBYTE)eztr.szHomeDir,
    &dwLen) != ERROR_SUCCESS)
  {
    return -2;
  }

  RegCloseKey(hKey);

  eztr.szHomeDir[OHB_HOME_DIR_LEN - 1] = '\0';

  len = lstrlenA(eztr.szHomeDir);
  if (len > 2000)
  {
    return -3;
  }

  lstrcatA(eztr.szHomeDir, "\\Dat");

  return 0;
}

static int OhbDestroyHomeDir(void)
{
  eztr.szHomeDir[0] = '\0';

  return 0;
}

int OhbEZTRJKInit(const OHBSETTING *pSetting)
{
  static int ret;

  if (bInit > 0)
  {
    ++bInit;
    return ret;
  }

  ++bInit;

  return (ret = OhbEZTRInit());
}

void OhbEZTRJKDestroy(void)
{
  --bInit;

  if (bInit > 0)
  {
    return;
  }

  OhbEZTRDestroy();
}

int OhbEZTRKJInit(const OHBSETTING *pSetting)
{
  return OhbEZTRJKInit(pSetting);
}

void OhbEZTRKJDestroy(void)
{
  OhbEZTRJKDestroy();
}

const OHBTRANS *OhbEZTRJKTrans(LPCWSTR jpStr, const OHBSETTING *pSetting)
{
  LPSTR szSrc;
  LPWSTR szDst;
  LPCSTR szTrans;
  int len;
  OHBTRANS *ohbTrans;

  OHBLOCK();

  len = WideCharToMultiByte(932, 0, jpStr, -1, 0, 0, 0, 0);
  szSrc = HeapAlloc(GetProcessHeap(), 0, len);
  WideCharToMultiByte(932, 0, jpStr, -1, szSrc, len, 0, 0);

  szTrans = eztr.J2K_TranslateMMNT(0, szSrc);

  len = MultiByteToWideChar(949, 0, szTrans, -1, 0, 0);
  szDst = HeapAlloc(GetProcessHeap(), 0, len * 2);
  MultiByteToWideChar(949, 0, szTrans, -1, szDst, len);

  ohbTrans = HeapAlloc(GetProcessHeap(), 0, sizeof(OHBTRANS));
  ohbTrans->pSetting = pSetting;
  ohbTrans->pData = (LPVOID)szTrans;
  ohbTrans->lpStr = szDst;
  HeapFree(GetProcessHeap(), 0, szSrc);

  OHBUNLOCK();

  return ohbTrans;
}

void OhbEZTRJKFree(const OHBTRANS *ohbTrans)
{
  OHBLOCK();
  HeapFree(GetProcessHeap(), 0, (LPVOID)ohbTrans->lpStr);
  eztr.J2K_FreeMem(ohbTrans->pData);
  HeapFree(GetProcessHeap(), 0, (LPVOID)ohbTrans);
  OHBUNLOCK();
}

const OHBTRANS *OhbEZTRKJTrans(LPCWSTR krStr, const OHBSETTING *pSetting)
{
  LPSTR szSrc;
  LPWSTR szDst;
  LPCSTR szTrans;
  int len;
  OHBTRANS *ohbTrans;

  OHBLOCK();

  len = WideCharToMultiByte(949, 0, krStr, -1, 0, 0, 0, 0);
  szSrc = HeapAlloc(GetProcessHeap(), 0, len);
  WideCharToMultiByte(949, 0, krStr, -1, szSrc, len, 0, 0);

  szTrans = eztr.K2J_TranslateMMNT(0, szSrc);

  len = MultiByteToWideChar(932, 0, szTrans, -1, 0, 0);
  szDst = HeapAlloc(GetProcessHeap(), 0, len * 2);
  MultiByteToWideChar(932, 0, szTrans, -1, szDst, len);

  ohbTrans = HeapAlloc(GetProcessHeap(), 0, sizeof(OHBTRANS));
  ohbTrans->pSetting = pSetting;
  ohbTrans->pData = (LPVOID)szTrans;
  ohbTrans->lpStr = szDst;
  HeapFree(GetProcessHeap(), 0, szSrc);

  OHBUNLOCK();

  return ohbTrans;
}

void OhbEZTRKJFree(const OHBTRANS *ohbTrans)
{
  OHBLOCK();
  HeapFree(GetProcessHeap(), 0, (LPVOID)ohbTrans->lpStr);
  eztr.K2J_FreeMem(ohbTrans->pData);
  HeapFree(GetProcessHeap(), 0, (LPVOID)ohbTrans);
  OHBUNLOCK();
}
